import json
import logging
import re
from os import environ
from pprint import pformat

from httmock import HTTMock, urlmatch
from wrapt import ObjectProxy


class NominodeContextMock(HTTMock):
    def __init__(self, task_parameters=None):
        super().__init__(self.api_match)
        self.logger = logging.getLogger("nomigen.nominode-mock")
        task_parameters = task_parameters or {}
        self.secrets = task_parameters.pop("config", {})
        self.params = task_parameters
        self.calls = []
        self.environ = None

    def __enter__(self):
        self.environ = environ.copy()
        environ["execution_uuid"] = "TEST_UUID"
        environ["task_uuid"] = "TASK_UUID"
        environ["project_uuid"] = "TEST_PROJECT"
        environ["nomnom_api"] = "http://127.0.0.1:9090"
        environ["token"] = "token"
        self.nnd_context = NominodeContext.from_env()
        self.nnd_context.__enter__()
        super().__enter__()
        return self

    def __exit__(self, *args):
        self.nnd_context.__exit__(*args)
        super().__exit__(*args)
        environ.pop("execution_uuid")
        environ.pop("task_uuid")
        environ.pop("project_uuid")
        environ.pop("nomnom_api")
        environ.pop("token")

    @urlmatch(netloc=r"(.*\.)?127.0.0.1:9090$")
    def api_match(self, url, request):
        self.calls.append((url.path, json.loads(request.body)))
        match = re.match(r"/connection/(?P<uuid>.+)/update", url.path)
        if match:
            json_data = request.body
            loaded = json.loads(json_data)
            uuid = match.groupdict()["uuid"]
            self.params["config"][uuid] = json.loads(loaded["parameters"])
            self.logger.debug("Caught connections update. Test creds updated")
        elif url.path == "/execution/log/TEST_UUID":
            json_data = request.body
            loaded = json.loads(json_data)
        elif url.path == "/task/TASK_UUID/update":
            json_data = request.body
            loaded = json.loads(json_data)
            self.logger.debug("Caught task update {}".format(pformat(loaded)))
        elif url.path == "/execution/update/TEST_UUID":
            json_data = request.body
            loaded = json.loads(json_data)
            self.logger.debug(
                "Caught execution progress update {}".format(pformat(loaded))
            )
        elif url.path == "/execution/decode/TEST_UUID":
            return json.dumps(self.secrets)
        elif url.path == "/task/TASK_UUID/parameters":
            json_data = request.body
            loaded = json.loads(json_data)
            self.logger.debug("Caught task parameter update {}".format(pformat(loaded)))
            return json.dumps({"result": "success"})
        elif url.path == "/execution/checkout/TEST_UUID":
            return json.dumps({"parameters": self.params, "task_uuid": "TASK_UUID"})
        else:
            self.logger.info(
                f"Unknown api endpoint called {url.path}, \n Body {request.body}"
            )
        return '{"you_logged":"test"}'


class NominodeContext:
    def __init__(self, execution_uuid, task_uuid, project_uuid, nomnom_api, token):
        self.execution_uuid = execution_uuid
        self.task_uuid = task_uuid
        self.project_uuid = project_uuid
        self.nomnom_api = nomnom_api
        self.token = token
        self.api_mock = NoContext()

    def __enter__(self):
        nominode_ctx._set_nominode_context(self)
        return self

    def __exit__(self, *args):
        nominode_ctx._set_nominode_context(NoContext())

    @classmethod
    def from_env(cls):
        try:
            instance = cls(
                execution_uuid=environ["execution_uuid"],
                task_uuid=environ["task_uuid"],
                project_uuid=environ["project_uuid"],
                nomnom_api=environ["nomnom_api"],
                token=environ["token"],
            )
        except KeyError as e:
            raise RuntimeError(
                f"Could not find expected environment variable '{e.args[0]}'"
            )
        return instance


class NoContext:
    def __getattr__(self, name):
        if name in ["execution_uuid", "task_uuid", "project_uuid", "nomnom_api", "token"]:
            raise RuntimeError(
                "Not in nominode context, wrap with ExecutionContextMock or only execute via a nominode"
            )
        else:
            return super().__getattribute__(name)


class NominodeContextProxy(ObjectProxy):
    def __init__(self):
        super().__init__(None)

    def _set_nominode_context(self, nominode_ctx):
        super().__init__(nominode_ctx)


nominode_ctx: NominodeContext = NominodeContextProxy()



import logging
from abc import ABCMeta, abstractmethod
from typing import List

from nomnomdata.engine.errors import MissingParameters, ValidationError

py_type = type

logger = logging.getLogger(__name__)


class ParameterType(metaclass=ABCMeta):
    """
    ParameterType is the base class for all parameter types.
    Should never be directly instantiated.
    """

    @property
    @abstractmethod
    def type(self):
        pass

    @property
    def shared_object_type_uuid(self):
        return None

    def validate(self, val):
        return True

    def dump(self, val):
        return val

    def load(self, val):
        return val


class Parameter:
    """
        Parameter for use in a ParameterGroup, also contains options
        to control the look in the Nominode UI

        :param type:
            Set the type to allow validation of the parameter and
            so the UI understands how to render this parameter.
            Example valid parameter type :class:`~nomnomdata.engine.parameters.String`
        :param name:
            Parameter name, will be the key used in the final result passed to
            your function.
        :param display_name:
            The name of the parameter to be displayed to the user.
        :param help_header_id:
            Header ID in any MD file declared in upper scope that will be
            linked to this parameter.
        :param help_md_path:
            Full path to an MD file that will be linked
            to this parameter. Cannot be used with help_header_id.
        :param required:
            Setting this to true will require the user to set a value for this parameter,
            defaults to False
        :param description:
            The long form description the UI will diplay next
            to the parameter
        :param default:
            Default value of the parameter will be set as on task creation,
            valid values vary by the ParameterType you use, defaults to nothing
    """

    def __init__(
        self,
        type: ParameterType,
        name: str,
        display_name: str = "",
        help_header_id: str = None,
        help_md_path: str = None,
        required: bool = False,
        description: str = "",
        default: object = None,
    ):
        if not isinstance(type, ParameterType) and issubclass(type, ParameterType):
            raise ValueError(f"{type} is not an instance, perhaps a forgotten ()?")
        elif not isinstance(type, ParameterType):
            raise ValueError(
                f"{type} is type {py_type(type)} when it must be an instance of a {ParameterType} subclass"
            )
        self.type = type
        self.name = name
        if help_header_id and help_md_path:
            raise ValueError("Cannot use both help_header_id and help_md_path")
        if help_header_id:
            self.help = {"header_id": help_header_id}
        elif help_md_path:
            self.help = {"file": help_md_path}
        else:
            self.help = None
        self.display_name = display_name or self.name.capitalize()
        self.required = required
        self.description = description
        self.default = default

    def validate(self, val):
        self._verify_required(val)
        if val is not None:
            return self.type.validate(val)
        else:
            return True

    def _verify_required(self, val):
        if self.required and val is None:
            raise MissingParameters()

    def __str__(self):
        return f"{self.type.__class__.__name__} Parameter"

    def __repr__(self):
        return f"<{self.type.__class__.__name__} Parameter>"


class ParameterGroup:
    """ParameterGroup acts as a logical and visual grouping for parameters

        :param name:
            Unique key for this group, defaults to "general"
        :warning: you cannot use parameter groups with the same name on one action
        :param display_name:
            Display name to render in the UI, defaults to "General Parameters"
        :param description: UI description for this parameter group, defaults to ""
        :param collapsed:
            If the UI should collapse this parameter group when rendering,
            defaults to False
        :param shared_parameter_group_uuid:
            UUID of the shared parameter group defined for this parameter,
            can be safely left undefined in 95% of cases
    """

    def __init__(
        self,
        *args: Parameter,
        name: str = "general",
        display_name: str = "General Parameters",
        description: str = "",
        collapsed: bool = False,
        shared_parameter_group_uuid: str = "",
    ):

        self.parameters = args
        self.collapsed = collapsed
        self.shared_parameter_group_uuid = shared_parameter_group_uuid
        self.description = description
        self.name = name
        self.display_name = display_name
        self.type = "group"


class Connection(ParameterType):
    """
        Connections are generally used for authentication credentials,
        parameters will be stored encrypted on the Nominode instead of in plain text and can require
        seperate permissions to be viewed or used.


        :param connection_type_uuid:
            Unique ID associated with this Connection
        :param parameter_groups:
            ParameterGroups in this connection
        :param description:
            Description of the connection to be rendered on the UI, defaults to ""
        :param alias:
            Short(er) alias for the connection
        :param categories:
            Categories this connection belongs to, for easier sorting in the UI.
    """

    type = "connection"

    def __init__(
        self,
        connection_type_uuid: str,
        parameter_groups: List[ParameterGroup],
        description: str = "",
        alias: str = "",
        categories: List[str] = None,
    ):

        self.connection_type_uuid = connection_type_uuid
        self.parameter_groups = parameter_groups
        self.description = description
        self.alias = alias
        self.categories = categories

    def load(self, value):
        all_keys = set(self.all_parameters.keys()) | set(value.keys())
        union_params = {
            key: (self.all_parameters.get(key), value.get(key)) for key in all_keys
        }
        kwargs = {}
        for keyword, (param, val) in union_params.items():
            kwargs[keyword] = param.type.load(val) if val is not None else val
        return kwargs

    def validate(self, value):
        if value:
            all_keys = set(self.all_parameters.keys()) | set(value.keys())
            union_params = {
                key: (self.all_parameters.get(key), value.get(key)) for key in all_keys
            }
            for keyword, (param, val) in union_params.items():
                try:
                    param.validate(val)
                except ValidationError as e:
                    e.add_key(keyword)
                    raise

    @property
    def all_parameters(self):
        return {p.name: p for pg in self.parameter_groups for p in pg.parameters}



from .base import Connection, Parameter, ParameterGroup, ParameterType
from .engine import Engine, current_engine

__all__ = [
    "Engine",
    "Parameter",
    "ParameterGroup",
    "ParameterType",
    "Connection",
    "current_engine",
]



import warnings

import click

from . import __version__


@click.group(name="engine")
@click.version_option(version=__version__, prog_name="nomnomdata-engine")
def cli():
    """NomNomData Engine CLI, used for the execution of engines"""
    warnings.warn(
        "nnd engine cli is deprecated, please use new python models", DeprecationWarning,
    )
